/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   mo_fedotov@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef AR_CLIENT_H
#define AR_CLIENT_H

#include "bt_comm.h"
#include "ar_parser.h"

#define ALARM_CLEAN 	0
#define ALARM_CHECK 	1

#define STR_MOTOROLA	"Motorola"
#define STR_SE   	"Sony Ericsson"
#define STR_SAGEM	"SAGEM"
#define STR_SIEMENS   	"SIEMENS"

#define MODEL_DEFAULT	0
#define MODEL_SE	1
#define MODEL_MOTOROLA	2
#define MODEL_SAGEM	3
#define MODEL_SIEMENS	4

/**
	@author M.Fedotov <mo_fedotov@mail.ru>
*/
class AR_Client {

public:
    AR_Client();
    ~AR_Client();

    bool connectToBT	    (QString s);
    int  listenForConnection(void);
    bool isConnected	    (void);
    bool initServerPort	    (void);
    void setStyle   	    (QString device);
    void setInitialized     (void);
    bool isServer	    (void);
    bool stillPressed	    (void);
    int  checkCallActive    (void);
    int  execCmd    	    (int type, QString command);
    type_key* getAutoRepeatCmdIndex(void);
    int  handleCmdByKey     (type_key *k);
    int  handleCommand	    (char * c);
    int  readCommand	    (QString& buffer, int l);
    void Disconnect 	    (void);
    void handleDisconnect   (bool needClose);
    void manageAlarms	    (int m);
    void setRemoteOn	    (bool set);
    int  processCommands    (cmdItem* cmdList);
    int processSendCmd	    (QString origCmd);
    	
    void freeTimers         (void);
    int  verifyTimers       (int ticksInSec);
    int  cancelTimer        (QString id);
    int  processTimerCmd    (QString origCmd);
    
    int     model;
    int     serverMode;
    int     port;
    int     initialized;
    timerCmd* timers;
      
private:

    int handleKeyPress	(char* key);
    int execSimpleCmd	(QString command, int toFile);
    int execInFork  	(const char *cmd);
    bool isBemusedCommand(char* cmd);
    
    QString  	readResultsFile();  
    int     	execCmdAndGetResults(int type, QString commandOrig, timerCmd *tm);
    
    timerCmd*   findTimer       (QString id, timerCmd **prev);
    int         createTimer     (QString id, int type, int tmout, int times, QString cmd);
	
    bool    callOn;
    bool    remoteOn;
    bool    repeatNow;
    bool    useCallId;
    BtComm  Comm;
    type_key*  repeatCmdPtr;
    QString callerId;
    QString modifierString;
    int     paramValue;
};

#endif
