/***************************************************************************
 *   Copyright (C) 2006 by M.Fedotov   *
 *   mo_fedotov@mail.ru   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qfileinfo.h>
#include <ksimpleconfig.h>

#include "kanyremote.h"
#include "ar_parser.h"
#include "logger.h"

//extern	AR_Parser *conf;
extern	logger2   *theLog;

int  	alarmTimer = 0;

///////////////////////////////////////////////////////////////////////////////////////

AR_Parser::AR_Parser()
{
    // Set defaults firstly
    autoConnect = false;        // AT mode only
    autoRepeat	= false;
    baudrate 	= 19200;
    charset 	= "8859-1";     // AT mode only
    device  	= DEFAULT_DEVICE;
    log		= false;
    debug   	= true;
    numAlarms	= 0;
    retrysecs	= 60;
    screen  	= true;
    modes	= NULL;
    defMode	= NULL;
    currentMode = NULL;
    aliases     = NULL;
}


AR_Parser::~AR_Parser()
{
}

bool AR_Parser::getLog()
{
    return log;
}

bool AR_Parser::getDebug()
{
    return debug;
}

void AR_Parser::setLog(bool newL)
{
    log = newL;
}

bool AR_Parser::getAutoConnect()
{
    return autoConnect;
}

void AR_Parser::setAutoConnect(bool newC)
{
    autoConnect = newC;
}

bool AR_Parser::getAutoRepeat()
{
    return autoRepeat;
}

QString AR_Parser::getDevice()
{
    return device;
}

void AR_Parser::setDevice(QString newD)
{
    device = newD;
}

int AR_Parser::getBaudrate()
{
    return baudrate;
}

int AR_Parser::getRetrySecs()
{
    return retrysecs;
}

bool AR_Parser::getUseScreen()
{
    return screen;
}

QString AR_Parser::getCharset()
{
    return charset;
}

QString AR_Parser::getToMainMenu()
{
    return  toMain;
}

QString AR_Parser::getAT_CMER(int what)
{
    if (what == CMER_ON)
    	return  cmerOn;
    if (what == CMER_OFF)
    	return  cmerOff;
    return QString::null;
}

int  AR_Parser::getNumAlarms()
{
    return numAlarms;
}

QString AR_Parser::getAlarmInfo(int which, cmdItem **cmds)
{
    if (which > numAlarms) return QString::null;
     
    *cmds = AlarmsData[which].cmd;
    return AlarmsData[which].alarmFile;
}

// ----------------- Setup configuration functions ------------------------- 

int AR_Parser::setBaudrate(int rate) 
{
  switch (rate) {
    case 300:
        return 300;
    case 1200:
        return 1200;
    case 2400:
        return 2400;
    case 9600:
        return 9600;
    case 19200:
        return 19200;
    case 38400:
        return 38400;
#ifdef B57600
    case 57600:
        return 57600;
#endif
#ifdef B115200
    case 115200:
        return 115200;
#endif
#ifdef B230400
    case 230400:
        return 230400;
#endif
    default:
        theLog->info("bad baudrate " + QString::number(rate) + " defaulting to 9600");
  }

  return 9600;
}

void AR_Parser::printParams()
{
    int atMode = 1;
    if (device.contains(INET_SOCKET) || device.contains(BT_SOCKET)) {
        atMode = 0;
    }
    
    
    theLog->cfg(QString("[Parameters]"));
    if (atMode) {
        theLog->cfg(QString("\tAuto Connect  ") + QString(autoConnect ? "Yes":"No"));
    }
    theLog->cfg(QString("\tAuto Repeat   ") + QString(autoRepeat ? "Yes":"No"));
    theLog->cfg(QString("\tBaudrate      ") + QString::number(baudrate));
    if (atMode) {
        if (cmerOn.isNull()) {
    	    theLog->cfg(QString("\tCMER On       Will be determined automatically"));
        } else {
    	    theLog->cfg(QString("\tCMER On       ") + cmerOn);
        }
        if (cmerOff.isNull()) {
    	    theLog->cfg(QString("\tCMER Off      Will be determined automatically"));
        } else {
    	    theLog->cfg(QString("\tCMER Off      ") + cmerOff);
        }
    }
    theLog->cfg(QString("\tDebug         ") + QString(debug ? "Yes":"No"));
    theLog->cfg(QString("\tDevice        ") + device);
    theLog->cfg(QString("\tLog           ") + QString(log ? "Yes":"No"));
    theLog->cfg(QString("\tRetry Sec     ") + QString::number(retrysecs));
    theLog->cfg(QString("\tScreen Menu   ") + QString(screen ? "Yes":"No"));
    if (atMode) {
        theLog->cfg(QString("\tTo Main Menu  ") + toMain);
    }
    theLog->cfg(QString("[End]"));
}

QString AR_Parser::id2Cmd (int cmdId) {
    switch(cmdId) {
        case ID_EXIT:               return QString(CMD_EXIT);
        case ID_TMPDISCONN:         return QString(CMD_TMPDISCONN);
        case ID_REMOTEON:           return QString(CMD_REMOTEON);
        case ID_REMOTEOFF:          return QString(CMD_REMOTEOFF);
        case ID_EXEC:               return QString(CMD_EXEC);
        case ID_SENDCKPD:           return QString(CMD_SENDCKPD);
        case ID_SETMODE:            return QString(CMD_SETMODE);
        case ID_SET:                return QString(CMD_SET);
        case ID_EXECSET:            return QString(CMD_EXECSET);
        case ID_TIMERCMD:           return QString(CMD_TIMER);
        case ID_SEND:               return QString(CMD_SEND);
        case ID_EXECSEND:           return QString(CMD_EXECSEND);
        case ID_CANCELTIMER:        return QString(CMD_CANCELTIMER);
        
        default:                    return QString("Unknown");
    }
}

void AR_Parser::printKeys()
{    
    mode *mp = modes;

    while (mp) {
        theLog->cfg(QString(MODE_STR) + " " + mp->name);
         
        type_key* It = mp->keys;
        QString tmp;
        
        while (It && It->key[0] != '\0') {
            if (It->cmd) {
            
                tmp = "\t" + It->key + " = ";
                
                cmdItem* item = It->cmd;
                while (item) { 
                    
                    tmp = tmp + id2Cmd(item->type);
                    if ( !(item->descr.isNull() || item->descr.isEmpty())) {
                        tmp = tmp + "(" + item->descr + ")";
                    }
                    tmp = tmp + ";";
                    
                    item = item->next;
                }
                
    	    } else {
                tmp = "\t" + It->key + "= no command";
            }    
    	    theLog->cfg(tmp);
	
	    It = (type_key*)It->next;
        }
        theLog->cfg(QString(MODE_END_STR) + " " + mp->name);
        
        mp = (mode*)mp->next;
    }
    
    return;
}

void AR_Parser::printAlarms()
{
    int i = 0;
    theLog->cfg(ALARMS_SECTION_STR);
    while (!(AlarmsData[i].alarmFile.isNull()) && i < numAlarms) {
    	QString tmp = "\t" + AlarmsData[i].alarmFile + ": ";

        cmdItem* item = AlarmsData[i].cmd;
        while (item) { 
            
            tmp = tmp + id2Cmd(item->type);
            if ( !(item->descr.isNull() || item->descr.isEmpty())) {
                tmp = tmp + "(" + item->descr + ")";
            }
            tmp = tmp + ";";
            
            item = item->next;
        }

    	theLog->cfg(tmp);
    	i++;
    }
    theLog->cfg(SECTION_END_STR);
}

void AR_Parser::printAliases()
{
    type_alias *al = aliases;
    theLog->cfg(ALIAS_SECTION_STR);
    while(al) {
    	theLog->cfg("\tKey " + al->key + " \taliased to " + al->alias);
        al = al->next;
    }
    theLog->cfg(SECTION_END_STR);
    return;
}

int AR_Parser::storeCmds(cmdItem** cmdList, QString value) {
    
    QString name;
    QString body;
    int     ptr;
    cmdItem *curCmd;
    int idx, cmdId, skipToSemicolon;
    
    
    if (value.isNull() || value.isEmpty()) {
        theLog->error("storeCmds(): null input\n"); 
        return EXIT_OK;
    }
    //theLog->debug("storeCmds() " + value); 
     
    curCmd = (cmdItem*) NULL;
    ptr = 0; // start of the string
    int len = value.length();
    
    while (ptr < len) {
    	
	int start = ptr;
    
        //theLog->debug("Got full cmd >" + value.mid(ptr) + "<");
        
        idx = 0;
        skipToSemicolon = 0;
        
        // Get name
        while(ptr < len && value.at(ptr) != '(' && value.at(ptr) != ';') {
            ptr++;
        }
        name = value.mid(start, ptr-start).simplifyWhiteSpace();  
        
        //theLog->debug("Got name >" + name + "<");
        
        if (name == CMD_EXIT) {
            cmdId = ID_EXIT;
            skipToSemicolon = 1;
        } else if (name == CMD_TMPDISCONN) {
            cmdId = ID_TMPDISCONN;
            skipToSemicolon = 1;
        } else if (name == CMD_REMOTEON) {
            cmdId = ID_REMOTEON;
            skipToSemicolon = 1;
        } else if (name == CMD_REMOTEOFF) {
            cmdId = ID_REMOTEOFF;
            skipToSemicolon = 1;
        } else if (name == CMD_EXEC) {
            cmdId = ID_EXEC;
        } else if (name == CMD_SENDCKPD) {
            cmdId = ID_SENDCKPD;
        } else if (name == CMD_SETMODE) {
            cmdId = ID_SETMODE;
        } else if (name == CMD_SET) {
            cmdId = ID_SET;
        } else if (name == CMD_EXECSET) {
            cmdId = ID_EXECSET;
        } else if (name == CMD_TIMER) {
            cmdId = ID_TIMERCMD;
        } else if (name == CMD_SEND) {
            cmdId = ID_SEND;
        } else if (name == CMD_EXECSEND) {
            cmdId = ID_EXECSEND;
        } else if (name == CMD_CANCELTIMER) {
            cmdId = ID_CANCELTIMER;
        } else {
            //theLog->debug(Unknown command name");
            //here could be params like log=... also
            cmdId = ID_UNKNOWN;
            skipToSemicolon = 1;
        }

	if (skipToSemicolon) {      // Skip until next ';'
            while (value.at(ptr) != '\0' && value.at(ptr) != ';') {
                ptr++;
            }
            if (value.at(ptr) != '\0') {
                 ptr++;
            }
            if (cmdId == ID_UNKNOWN) {
                continue;               // Next loop - read command by command
            }
        }
     
        // skip possible spaces
        while (ptr < len && (value.at(ptr) == ' ' || value.at(ptr) == '\t')) {
            ptr++;
        }
    
        // We've got name of command, now we have to get body
        if (skipToSemicolon) {  	// Command without body 
            body = QString::null;
        } else {
            
            while (ptr < len && value.at(ptr) != '(') {
                ptr++;
            }
            if (value.at(ptr) == '(') {
                ptr++;
            }
            int braces = 1;         // cmd_name(          
            idx = ptr;              // start position of a body
            
            while (ptr < len && !(value.at(ptr) != ';' && braces == 0)) {
                if (value.at(ptr) == '(') {
                    braces++;
                }
                if (value.at(ptr) == ')') {
                    braces--;
                }
                ptr++;
            } 
            int ptrB = ptr;
            while (ptrB > 0 && value.at(ptrB) != ')') {    // go backward to strip last ')' 
                ptrB--;
            }
            if (value.at(ptrB) == ')') {
                ptrB--;
            }
            body = value.mid(idx,ptrB-idx+1);
        }
        //theLog->debug("Got command body >" + body + "<");
        
        // Insert into command list
        cmdItem* newCmd = new cmdItem();
        
        newCmd->type = cmdId;
        newCmd->next = NULL;
        if (body.isEmpty() || body.isNull()) {
            newCmd->descr = QString::null;
        } else {
            newCmd->descr = body;
        }
        
        if (curCmd == NULL) { // Insert first
            *cmdList = newCmd;
        } else {
            curCmd->next = newCmd;
        }
        curCmd = newCmd;
        
    }   // while
    
    return EXIT_OK;
}

void AR_Parser::storeKey(type_key** head, QString tag, QString value) 
{
    type_key* It = NULL;
    int ret;
    //theLog->debug("storeKey() " + tag);
    
    It = new type_key();

    // Insert in head
    if ((*head) == NULL) {   // first elem
        It->next = NULL;
    } else {
        It->next = (type_key*) (*head);
    }
    (*head) = It;

    It->key = tag.simplifyWhiteSpace();     	// Make items separated only by one space
    if (value.isNull() || value.isEmpty()) {
    	It->cmd = NULL;
    } else {
    	ret = storeCmds(&(It->cmd), value);
    }

    return;
}

void AR_Parser::loadKeys(QTextStream& stream) 
{
    mode *curMode = modes;

    while (true) {
        QString keyDef = stream.readLine();
    	//theLog->debug("loadKeys " + keyDef);
            
    	if (keyDef.at(0) == QChar('%')) {    // commented out line
    	    continue;
    	}
        if (keyDef.isNull()) {
            return;
        }
        
        if (keyDef.contains(QString("[End]"))) {
            return;
        } else if (keyDef.contains(QString(MODE_STR))){             // mode definition start
        
            mode *nm = (mode*) new mode();
            nm->name = (keyDef.section("=", 1, 1)).stripWhiteSpace();
            nm->keys = NULL;
            
            //theLog->debug("Mode = " + nm->name);
            
            // Insert in head
            if (modes == NULL) {   // first elem
                nm->next = NULL;
            } else {
                nm->next = (mode*) modes;
            }
            modes   = nm;
            curMode = nm;
        } else if(keyDef.contains(QString(MODE_END_STR))) {         // mode definition end
            curMode = defMode;
            //theLog->debug("EndMode");
        } else {
            QString keyStr = (keyDef.section("=", 0, 0)).stripWhiteSpace();
            QString cmdStr = (keyDef.section("=", 1)).stripWhiteSpace();    // All sections after first  =
            
	    //theLog->debug("Got " + keyStr + "=" + cmdStr);            

            if (keyStr.isNull() || keyStr.isEmpty()) {
                continue;
            } else {
                storeKey(&(curMode->keys), keyStr, cmdStr);
            }
        }
    }
}

void AR_Parser::loadAliases(QTextStream& stream)
{
    while (true) {
        QString line = stream.readLine();
            
    	if (line.at(0) == QChar('%')) {    // commented out line
    	    continue;
    	}
        if (line.isNull()) {
            return;
        }

        if (line.contains(QString("[End]"))) {
            return;
        } else {
            QString key   = (line.section("=", 0, 0)).stripWhiteSpace();
            QString alias = (line.section("=", 1)).stripWhiteSpace();
      
            if (key.isNull() || alias.isNull()) {
	    	continue;
            } else {
               type_alias* Al = new type_alias();

                // Insert in head
                if (aliases == NULL) {   // first elem
                    Al->next = NULL;
                } else {
                    Al->next = (type_alias*) aliases;
                }
                aliases = Al;

                Al->key   = key;
                Al->alias = alias;
            }
        }
    }

}
void AR_Parser::loadAlarms(QTextStream& stream)
{
    numAlarms = 0;
    
    while (true) {
        QString line = stream.readLine();
            
    	if (line.at(0) == QChar('%')) {    // commented out line
    	    continue;
    	}
        if (line.isNull()) {
            return;
        }

        if (line.contains(QString("[End]"))) {
            return;
        } else {
            QString file  = (line.section("=", 0, 0)).stripWhiteSpace();
            QString value = (line.section("=", 1)).stripWhiteSpace();
      
            if (file.isNull() || value.isNull()) {
	    	continue;
            } else {
            
                AlarmsData[numAlarms].alarmFile = file;
 		if (value == NULL) {
		    AlarmsData[numAlarms].cmd = NULL;
		} else {
		    storeCmds(&(AlarmsData[numAlarms].cmd), value);
		}
               
                numAlarms++;
    	    
                if (numAlarms == MAXALARMS) {
    	    	    theLog->cfg("No more room for alarms");
    	    	    return;
    	        }
            }
        }
    }
}

int AR_Parser::load_cfg(QString mfile)
{
    
    if (!(QFile(mfile)).exists()) {
    	return EXIT_NOK;
    }
	
    KSimpleConfig theConfig(mfile);

    // String
    toMain    = theConfig.readEntry(TOMAIN_STR);
    cmerOn    = theConfig.readEntry(CMER_ON_STR);
    cmerOff   = theConfig.readEntry(CMER_OFF_STR);
    device    = theConfig.readEntry(DEVICE_STR);
    charset   = theConfig.readEntry(CHARSET_STR);
	
    // Numeric
    retrysecs = theConfig.readNumEntry(RETRY_STR,	     retrysecs);
    baudrate  = setBaudrate(theConfig.readNumEntry(BAUD_STR, baudrate));

    // Boolean
    log	    	= theConfig.readBoolEntry(LOG_STR,  	    log);
    debug	= theConfig.readBoolEntry(DEBUG_STR,  	    debug);
    screen  	= theConfig.readBoolEntry(SCREEN_STR,	    screen);
    autoConnect = theConfig.readBoolEntry(AUTOCONN_STR,     autoConnect);
    autoRepeat  = theConfig.readBoolEntry(AUTOREPEAT_STR,   autoRepeat);

    // We have at least default mode
    defMode = new mode();
    defMode->name ="default";
    defMode->next = NULL;
    defMode->keys = NULL;
    	    
    // Add it to the modes list
    modes         = defMode;
    currentMode   = defMode;
    
    // Unfortunately, KSimpleConfig does not allow to '#', '[' to be the key elements (<key>=<value>)
    // So, it needs to do this dirty job by hands
    
    QFile file(mfile);
    
    if (file.open(IO_ReadOnly)) {
    	QTextStream stream( &file );
        
	QString text;
        
     	do {
    	    text = stream.readLine();
		
    	    if (text.isNull()) {
    	    	break;
    	    }
	    if (text.at(0) == QChar('%')) {    // commented out line
    	    	continue;
    	    }
    	    	    
    	    if (text.contains(KEYS_SECTION_STR)) {
     	        loadKeys(stream);
    	    } else if (text.contains(ALIAS_SECTION_STR)) {
                loadAliases(stream);
    	    } else if (text.contains(ALARMS_SECTION_STR)) {
                loadAlarms(stream);
            }
    	} while (!text.isNull());
    	file.close();
    } else {
       	return EXIT_NOK;
    }

    // Thanks to KSimpleConfig once more...

    return EXIT_OK;
}

void AR_Parser::printConf() 
{
    if (!this->getLog()) return;

    printKeys 	();
    printAliases();
    printAlarms ();
    printParams ();

    theLog->cfg("Configuration printed out ...");

    return ;
}


int AR_Parser::load(QString path)
{
    //theLog->debug("AR_Parser::load");

    if (path.isNull() || path.isEmpty()) {
    	QString cfgfile = getenv("HOME") + QString("/") + CFGFILE;
	
    	theLog->info("Use cfg file >" + cfgfile + "<");

    	if(load_cfg(cfgfile) != EXIT_OK) {
            // Try again in current dir 
            cfgfile = QString("./") + CFGFILE;
    	    theLog->info("Try to use cfg file >" + cfgfile + "<");
    	    
	    return load_cfg(cfgfile);
    	}
    } else {
    	// Make path absolute
        if (path.at(0) != QChar('/')) {
	    //theLog->debug("Make abs. path");
	    path = (QFileInfo(path)).absFilePath(); 
	}
    	theLog->info("Use cfg file >" + path + "<");
    	return load_cfg(path);
    }
    
    return EXIT_NOK;
}

//------------------------- Utility finctions -------------------------- 

QString AR_Parser::findAlias(QString keyIn)
{
    //theLog->debug("findAlias");
    type_alias* al = aliases;
    QString key;

    if (keyIn.isNull() || keyIn.isEmpty()) {
    	theLog->info("findAlias: Got empty key");   
    	key = EMPTY_STR;
    } else {
    	key = keyIn;
    }
    
    while (al) {
        //sprintf(tmp,"compare alias >%s< = >%s<", key, al->key);
        //logger("DEBUG",tmp);   
    
        if (strcmp(key, al->key) == 0) {
     	    theLog->debug("found alias " + key + "=" + al->alias);
       
            return al->alias;
        }
        al = al->next;
    }
    return QString::null;
}

type_key* AR_Parser::findItemInMode(mode *mode, QString key, int& flag) 
{
    if (mode == NULL) {
    	theLog->debug("findItemInMode: mode is null");
    	return NULL;
    }

    //theLog->debug("findItem >" + mode->name + " " + key + "<");   

    type_key* it = mode->keys;
    
    while (it && (! it->key.isNull()) && it->key != key) {
    	//theLog->debug("key=" + it->key);
	it = it->next; 
    }
    if (it && (!(it->key).isNull())) {
    	//theLog->debug("!key.isNull()");
	flag = 1;
    	return it;
    } 
    
    it = mode->keys;
    while (it && (!it->key.isNull())) {
    
    	//theLog->debug("key=" + it->key);
	
    	int start = it->key.find(key);
    
    	if (start == 0 && (!it->key.isNull())) { // Index should be 0
    	    // We've got a part of multi key command
	    //theLog->debug("found part of multi key seq");
	    flag = 0;
    	    return it;
    	} 
        it = it->next;
    }
    
    theLog->debug("findItem: Item does not found for key=" + key);
    return NULL;
}

type_key* AR_Parser::findItem(QString keyIn, int& flag) 
{
    type_key *tk = NULL;
    
    //theLog->debug("findItem()");  

    QString key;
    
    if (keyIn.isEmpty()) {
    	theLog->debug("Got empty CKEV");   
    	key = EMPTY_STR;
    } else {
    	key = keyIn;
    }
    if (currentMode != NULL) {
    	tk = findItemInMode(currentMode, key, flag);
    	if (tk == NULL && currentMode != defMode) {
    	    theLog->debug("findItem() search in default");
    	    tk = findItemInMode(defMode, key, flag);
    	}
    } else {
    	theLog->error("Current mode is not defined!");
    }
    
    return tk;
}

cmdItem* AR_Parser::getCommand(type_key *it)
{
    if (it == NULL) {
     	theLog->debug("getCommand() NULL input");
	return NULL;
    }
    if (it->key.isNull()) {
     	theLog->debug("getCommand() NULL key");
	return NULL;
    }
    
    theLog->debug("getCommand for >" + it->key + "<");
    
    return it->cmd;
}

QString AR_Parser::parse_CLCC(QString clcc)
{
    // There are could be more than one call at time ... but here we will get then one-by-one
    // +CLCC: 1,,4,0,0,"<phone number here>",129,
    QString ID;
    
    QString ptr = clcc.section(',',5,5);
    
    if (ptr.isNull() || ptr.isEmpty()) {    // Call was finished
    	return QString("FINISHED");
    }
    
    if (ptr.at(0) == QChar('"')) {
    	ID = ptr.mid(1, ptr.length() - 2);
    } else {
    	ID = ptr;
    }
    
    if(ID.isNull() || ID.isEmpty()) {	    // Empty CallerID
    	theLog->debug("parse_CLCC >NULL callerID<");
	return QString("NO CALLER ID");
    } else {
    	theLog->debug(QString("parse_CLCC >") + ID + "<");
    }
    
    return ID;
}

void  AR_Parser::setMode (QString modeName) 
{
    if (modeName.isNull() || modeName.isEmpty()) {
        theLog->debug("setMode() NULL input");
        return;
    }
    theLog->debug("setMode() to " + modeName);

    mode *mp = modes;
    while (mp && (!modeName.contains(mp->name))) {
        mp = (mode*) mp->next;
    }
    
    if (mp) {
        currentMode = mp;
        theLog->debug("setMode() new mode was set to " + currentMode->name);
    } else {
        theLog->debug("setMode() new mode did not found");
    }
    
    return;
}

QString AR_Parser::getMode (void) 
{
    return currentMode->name;
}

void  AR_Parser::freeCfg(void) 
{
    theLog->debug("freeCfg()");
    
    // Keys
    mode *modesTmp = modes;
    while(modesTmp) {
        mode     *md = modesTmp;
        type_key *mk = md->keys;
        
        while (mk) {
    	    type_key* It = mk;
            
            cmdItem* item = It->cmd;
            cmdItem* mi;
            while (item) {
                mi = item->next;
                delete item;
                item = mi;
            }

	    mk = (type_key*) mk->next;
	    delete It;
        }
        modesTmp = (mode*) modesTmp->next;
        delete md;
    }
    modes = NULL;

    // Alarms
    int numAlarms = getNumAlarms();
    int i = 0;
    while (i<numAlarms) {

        cmdItem* item= AlarmsData[i].cmd;
        cmdItem* mi;
        while (item) {
            mi = item->next;
            delete item;
            item = mi;
        }
	AlarmsData[i].cmd = NULL;
	
	i++;
    }

    // Aliases
    type_alias *al = aliases;
    while(al) {
        type_alias* al2 = al;
        al = (type_alias*) al->next;
        delete al2;
    }
    aliases = NULL;
    
    theLog->debug("freeCfg() OK");

    return;
}
